"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from .BoldModel import BoldModel

#####################################################################
### Several balloon model variants following Stephan et al. (2007)
###
### C = classical coefficient (e. g. no epsilon influence on k_2)
### R = revised coefficient (e. g. epsilon influence on k_2)
###
### N = non-linear BOLD equation
### L = linear bold equation
###
### Parameters:
### V_0, tau_0, tau_s, tau_f they don't mention --> use Friston et al. (2000)
### E_0, epsilon are free --> E_0 use Friston et al. (2000), epsilon use 1.43 (probably used if 1.43 is fixed in Stephan et al. (2007), value from Obata et al. (2004))
###
### References:
### Friston et al. (2000). Nonlinear responses in fmri: the balloon model, volterra kernels, and other hemodynamics.NeuroImage12, 466–477
### Obata et al. (2004).  Discrepancies between bold and flow dynamics in primary and supplementary motor areas: application of the balloon model to the interpretation of bold transients. NeuroImage 21, 144–153
#####################################################################
class balloon_CN(BoldModel):
    """
    A balloon model with classic coefficients and non-linear BOLD equation derived from Stephan et al. (2007).

    Equivalent code:

    ```python
    balloon_CN = BoldModel(
        parameters = dict(
            second    = 1000.0,
            phi       = 1.0,
            kappa     = 1/1.54,
            gamma     = 1/2.46,
            E_0       = 0.34,
            tau       = 0.98,
            alpha     = 0.33,
            V_0       = 0.02,
            v_0       = 40.3,
            TE        = 40/1000.,
            epsilon   = 1.43,
        ),
        equations = [
            # Single input
            ann.Variable('I_CBF = sum(I_CBF)', init=0.0),
            ann.Variable('ds/dt = (phi * I_CBF - kappa * s - gamma * (f_in - 1))/second', init=0.0),
            ann.Variable('df_in/dt = s / second', init=1.0, min=0.01),

            ann.Variable('E = 1 - (1 - E_0)**(1 / f_in)', init=0.3424),
            ann.Variable('dq/dt = (f_in * E / E_0 - (q / v) * f_out)/(tau*second)', init=1.0, min=0.01),
            ann.Variable('dv/dt = (f_in - f_out)/(tau*second), init=1.0, min=0.01),
            ann.Variable('f_out = v**(1 / alpha)', init=1, min=0.01),

            # Classic coefficients
            ann.Variable('k_1 = (1 - V_0) * 4.3 * v_0 * E_0 * TE'),
            ann.Variable('k_2 = 2 * E_0'),
            ann.Variable('k_3 = 1 - epsilon'),

            # Non-linear equation
            ann.Variable('BOLD = V_0 * (k_1 * (1 - q) + k_2 * (1 - (q / v)) + k_3 * (1 - v))'),
        ],
        inputs="I_CBF",
    )
    ```

    :param phi:       input coefficient
    :param kappa:     signal decay
    :param gamma:     feedback regulation
    :param E_0:       oxygen extraction fraction at rest
    :param tau:       time constant (in s!)
    :param alpha:     vessel stiffness
    :param V_0:       resting venous blood volume fraction
    :param v_0:       frequency offset at the outer surface of the magnetized vessel for fully deoxygenated blood at 1.5 T
    :param TE:        echo time
    :param epsilon:   ratio of intra- and extravascular signal
    """
    def __init__(self,
            phi       = 1.0,
            kappa     = 1/1.54,
            gamma     = 1/2.46,
            E_0       = 0.34,
            tau       = 0.98,
            alpha     = 0.33,
            V_0       = 0.02,
            v_0       = 40.3,
            TE        = 40/1000.,
            epsilon   = 1.43,
        ):

        parameters = f"""
            second    = 1000.0 : population
            phi       = {phi} : population
            kappa     = {kappa} : population
            gamma     = {gamma} : population
            E_0       = {E_0} : population
            tau       = {tau} : population
            alpha     = {alpha} : population
            V_0       = {V_0} : population
            v_0       = {v_0} : population
            TE        = {TE} : population
            epsilon   = {epsilon} : population
        """ 

        equations = """
            # Single input
            I_CBF          = sum(I_CBF)                                                : init=0
            ds/dt          = (phi * I_CBF - kappa * s - gamma * (f_in - 1))/second     : init=0
            df_in/dt       = s / second                                                : init=1, min=0.01

            E              = 1 - (1 - E_0)**(1 / f_in)                                 : init=0.3424
            dq/dt          = (f_in * E / E_0 - (q / v) * f_out)/(tau*second)           : init=1, min=0.01
            dv/dt          = (f_in - f_out)/(tau*second)                               : init=1, min=0.01
            f_out          = v**(1 / alpha)                                            : init=1, min=0.01

            # Classic coefficients
            k_1            = (1 - V_0) * 4.3 * v_0 * E_0 * TE
            k_2            = 2 * E_0
            k_3            = 1 - epsilon

            # Non-linear equation
            BOLD           = V_0 * (k_1 * (1 - q) + k_2 * (1 - (q / v)) + k_3 * (1 - v))  : init=0
        """
        name = "BoldNeuron CN",
        description = "BOLD computation with classic coefficients and non-linear BOLD equation (Stephan et al., 2007)."

        BoldModel.__init__(self, 
            parameters=parameters, 
            equations=equations,  
            inputs="I_CBF",
            output="BOLD",
            name=name, description=description
        )
        

class balloon_CL(BoldModel):
    """
    A balloon model with classical coefficients and linear BOLD equation derived from Stephan et al. (2007).

    Equivalent code:

    ```python
    balloon_CL = BoldModel(
        parameters = dict(
            second    = 1000.0,
            phi       = 1.0,
            kappa     = 1/1.54,
            gamma     = 1/2.46,
            E_0       = 0.34,
            tau       = 0.98,
            alpha     = 0.33,
            V_0       = 0.02,
            v_0       = 40.3,
            TE        = 40/1000.,
            epsilon   = 1.43,
        ),
        equations = [
            # Single input
            ann.Variable('I_CBF = sum(I_CBF)', init=0.0),
            ann.Variable('ds/dt = (phi * I_CBF - kappa * s - gamma * (f_in - 1))/second', init=0.0),
            ann.Variable('df_in/dt = s / second', init=1.0, min=0.01),

            ann.Variable('E = 1 - (1 - E_0)**(1 / f_in)', init=0.3424),
            ann.Variable('dq/dt = (f_in * E / E_0 - (q / v) * f_out)/(tau*second)', init=1.0, min=0.01),
            ann.Variable('dv/dt = (f_in - f_out)/(tau*second), init=1.0, min=0.01),
            ann.Variable('f_out = v**(1 / alpha)', init=1, min=0.01),

            # Classic coefficients
            ann.Variable('k_1 = (1 - V_0) * 4.3 * v_0 * E_0 * TE'),
            ann.Variable('k_2 = 2 * E_0'),
            ann.Variable('k_3 = 1 - epsilon'),

            # Non-linear equation
            ann.Variable('BOLD = V_0 * ((k_1 + k_2) * (1 - q) + (k_3 - k_2) * (1 - v))'),
        ],
        inputs="I_CBF",
    )
    ```

    :param phi:       input coefficient
    :param kappa:     signal decay
    :param gamma:     feedback regulation
    :param E_0:       oxygen extraction fraction at rest
    :param tau:       time constant (in s!)
    :param alpha:     vessel stiffness
    :param V_0:       resting venous blood volume fraction
    :param v_0:       frequency offset at the outer surface of the magnetized vessel for fully deoxygenated blood at 1.5 T
    :param TE:        echo time
    :param epsilon:   ratio of intra- and extravascular signal
    """
    def __init__(self,
            phi       = 1.0,
            kappa     = 1/1.54,
            gamma     = 1/2.46,
            E_0       = 0.34,
            tau       = 0.98,
            alpha     = 0.33,
            V_0       = 0.02,
            v_0       = 40.3,
            TE        = 40/1000.,
            epsilon   = 1.43,
        ):

        parameters = f"""
            second    = 1000.0 : population
            phi       = {phi} : population
            kappa     = {kappa} : population
            gamma     = {gamma} : population
            E_0       = {E_0} : population
            tau       = {tau} : population
            alpha     = {alpha} : population
            V_0       = {V_0} : population
            v_0       = {v_0} : population
            TE        = {TE} : population
            epsilon   = {epsilon} : population
        """ 

        equations = """
            # Single input
            I_CBF          = sum(I_CBF)                                                : init=0
            ds/dt          = (phi * I_CBF - kappa * s - gamma * (f_in - 1))/second     : init=0
            df_in/dt       = s / second                                                : init=1, min=0.01

            E              = 1 - (1 - E_0)**(1 / f_in)                                 : init=0.3424
            dq/dt          = (f_in * E / E_0 - (q / v) * f_out)/(tau*second)           : init=1, min=0.01
            dv/dt          = (f_in - f_out)/(tau*second)                               : init=1, min=0.01
            f_out          = v**(1 / alpha)                                            : init=1, min=0.01

            # Classic coefficients
            k_1            = (1 - V_0) * 4.3 * v_0 * E_0 * TE
            k_2            = 2 * E_0
            k_3            = 1 - epsilon

            # Linear equation
            BOLD           = V_0 * ((k_1 + k_2) * (1 - q) + (k_3 - k_2) * (1 - v))        : init=0
        """
        name = "BoldNeuron CL"
        description = "BOLD computation with classic coefficients and linear BOLD equation (Stephan et al., 2007)."

        BoldModel.__init__(self, 
            parameters=parameters, 
            equations=equations,  
            inputs="I_CBF",
            output="BOLD",
            name=name, description=description
        )

class balloon_RN(BoldModel):
    """
    A balloon model with revised coefficients and non-linear BOLD equation derived from Stephan et al. (2007).

    Equivalent code:

    ```python
    balloon_RN = BoldModel(
        parameters = dict(
            second    = 1000.0,
            phi       = 1.0,
            kappa     = 1/1.54,
            gamma     = 1/2.46,
            E_0       = 0.34,
            tau       = 0.98,
            alpha     = 0.33,
            V_0       = 0.02,
            v_0       = 40.3,
            TE        = 40/1000.,
            epsilon   = 1.43,
            r_0       = 25.,
        ),
        equations = [
            # Single input
            ann.Variable('I_CBF = sum(I_CBF)', init=0.0),
            ann.Variable('ds/dt = (phi * I_CBF - kappa * s - gamma * (f_in - 1))/second', init=0.0),
            ann.Variable('df_in/dt = s / second', init=1.0, min=0.01),

            ann.Variable('E = 1 - (1 - E_0)**(1 / f_in)', init=0.3424),
            ann.Variable('dq/dt = (f_in * E / E_0 - (q / v) * f_out)/(tau*second)', init=1.0, min=0.01),
            ann.Variable('dv/dt = (f_in - f_out)/(tau*second), init=1.0, min=0.01),
            ann.Variable('f_out = v**(1 / alpha)', init=1, min=0.01),

            # Revised coefficients
            ann.Variable('k_1 = 4.3 * v_0 * E_0 * TE'),
            ann.Variable('k_2 = epsilon * r_0 * E_0 * TE'),
            ann.Variable('k_3 = 1 - epsilon'),

            # Non-linear equation
            ann.Variable('BOLD = V_0 * (k_1 * (1 - q) + k_2 * (1 - (q / v)) + k_3 * (1 - v))'),
        ],
        inputs="I_CBF",
    )
    ```

    :param phi:       input coefficient
    :param kappa:     signal decay
    :param gamma:     feedback regulation
    :param E_0:       oxygen extraction fraction at rest
    :param tau:       time constant (in s!)
    :param alpha:     vessel stiffness
    :param V_0:       resting venous blood volume fraction
    :param v_0:       frequency offset at the outer surface of the magnetized vessel for fully deoxygenated blood at 1.5 T
    :param TE:        echo time
    :param epsilon:   ratio of intra- and extravascular signal
    :param r_0:       slope of the relation between the intravascular relaxation rate and oxygen saturation
    """
    def __init__(self,
            phi       = 1.0,
            kappa     = 1/1.54,
            gamma     = 1/2.46,
            E_0       = 0.34,
            tau       = 0.98,
            alpha     = 0.33,
            V_0       = 0.02,
            v_0       = 40.3,
            TE        = 40/1000.,
            epsilon   = 1.43,
            r_0       = 25,
        ):

        parameters = f"""
            second    = 1000.0 : population
            phi       = {phi} : population
            kappa     = {kappa} : population
            gamma     = {gamma} : population
            E_0       = {E_0} : population
            tau       = {tau} : population
            alpha     = {alpha} : population
            V_0       = {V_0} : population
            v_0       = {v_0} : population
            TE        = {TE} : population
            epsilon   = {epsilon} : population
            r_0       = {r_0} : population
        """

        equations = """
            # Single input
            I_CBF          = sum(I_CBF)                                                : init=0
            ds/dt          = (phi * I_CBF - kappa * s - gamma * (f_in - 1))/second     : init=0
            df_in/dt       = s / second                                                : init=1, min=0.01

            E              = 1 - (1 - E_0)**(1 / f_in)                                 : init=0.3424
            dq/dt          = (f_in * E / E_0 - (q / v) * f_out)/(tau*second)           : init=1, min=0.01
            dv/dt          = (f_in - f_out)/(tau*second)                               : init=1, min=0.01
            f_out          = v**(1 / alpha)                                            : init=1, min=0.01

            # Revised coefficients
            k_1            = 4.3 * v_0 * E_0 * TE
            k_2            = epsilon * r_0 * E_0 * TE
            k_3            = 1.0 - epsilon

            # Non-linear equation
            BOLD           = V_0 * (k_1 * (1 - q) + k_2 * (1 - (q / v)) + k_3 * (1 - v))  : init=0
        """
        name = "BOLD model RN"
        description = "BOLD computation with revised coefficients and non-linear BOLD equation (Stephan et al., 2007)."

        BoldModel.__init__(self, 
            parameters=parameters, 
            equations=equations,  
            inputs="I_CBF",
            output="BOLD",
            name=name, description=description
        )

class balloon_RL(BoldModel):
    """
    A balloon model with revised coefficients and linear BOLD equation derived from Stephan et al. (2007).

    Equivalent code:

    ```python
    balloon_RL = BoldModel(
        parameters = dict(
            second    = 1000.0,
            phi       = 1.0,
            kappa     = 1/1.54,
            gamma     = 1/2.46,
            E_0       = 0.34,
            tau       = 0.98,
            alpha     = 0.33,
            V_0       = 0.02,
            v_0       = 40.3,
            TE        = 40/1000.,
            epsilon   = 1.43,
            r_0       = 25.,
        ),
        equations = [
            # Single input
            ann.Variable('I_CBF = sum(I_CBF)', init=0.0),
            ann.Variable('ds/dt = (phi * I_CBF - kappa * s - gamma * (f_in - 1))/second', init=0.0),
            ann.Variable('df_in/dt = s / second', init=1.0, min=0.01),

            ann.Variable('E = 1 - (1 - E_0)**(1 / f_in)', init=0.3424),
            ann.Variable('dq/dt = (f_in * E / E_0 - (q / v) * f_out)/(tau*second)', init=1.0, min=0.01),
            ann.Variable('dv/dt = (f_in - f_out)/(tau*second), init=1.0, min=0.01),
            ann.Variable('f_out = v**(1 / alpha)', init=1, min=0.01),

            # Revised coefficients
            ann.Variable('k_1 = 4.3 * v_0 * E_0 * TE'),
            ann.Variable('k_2 = epsilon * r_0 * E_0 * TE'),
            ann.Variable('k_3 = 1 - epsilon'),

            # Linear equation
            ann.Variable('BOLD = V_0 * ((k_1 + k_2) * (1 - q) + (k_3 - k_2) * (1 - v))'),
        ],
        inputs="I_CBF",
    )
    ```

    :param phi:       input coefficient
    :param kappa:     signal decay
    :param gamma:     feedback regulation
    :param E_0:       oxygen extraction fraction at rest
    :param tau:       time constant (in s!)
    :param alpha:     vessel stiffness
    :param V_0:       resting venous blood volume fraction
    :param v_0:       frequency offset at the outer surface of the magnetized vessel for fully deoxygenated blood at 1.5 T
    :param TE:        echo time
    :param epsilon:   ratio of intra- and extravascular signal
    :param r_0:       slope of the relation between the intravascular relaxation rate and oxygen saturation

    """
    def __init__(self,
            phi       = 1.0,
            kappa     = 1/1.54,
            gamma     = 1/2.46,
            E_0       = 0.34,
            tau       = 0.98,
            alpha     = 0.33,
            V_0       = 0.02,
            v_0       = 40.3,
            TE        = 40/1000.,
            epsilon   = 1.43,
            r_0       = 25,
        ):

        parameters = f"""
            second    = 1000.0 : population
            phi       = {phi} : population
            kappa     = {kappa} : population
            gamma     = {gamma} : population
            E_0       = {E_0} : population
            tau       = {tau} : population
            alpha     = {alpha} : population
            V_0       = {V_0} : population
            v_0       = {v_0} : population
            TE        = {TE} : population
            epsilon   = {epsilon} : population
            r_0       = {r_0} : population
        """

        equations = """
            # Single input
            I_CBF          = sum(I_CBF)                                                    : init=0
            ds/dt          = (phi * I_CBF - kappa * s - gamma * (f_in - 1))/second         : init=0
            df_in/dt       = s  / second                                                   : init=1, min=0.01

            E              = 1 - (1 - E_0)**(1 / f_in)                                     : init=0.3424
            tau*dq/dt      = f_in * E / E_0 - (q / v) * f_out                              : init=1, min=0.01
            tau*dv/dt      = f_in - f_out                                                  : init=1, min=0.01
            f_out          = v**(1 / alpha)                                                : init=1, min=0.01

            # Revised coefficients
            k_1            = 4.3 * v_0 * E_0 * TE
            k_2            = epsilon * r_0 * E_0 * TE
            k_3            = 1 - epsilon

            # Linear equation
            BOLD           = V_0 * ((k_1 + k_2) * (1 - q) + (k_3 - k_2) * (1 - v))         : init=0
        """
        name = "BOLD model RL"
        description = "BOLD computation with revised coefficients and linear BOLD equation (Stephan et al., 2007)."

        BoldModel.__init__(self, 
            parameters=parameters, 
            equations=equations,  
            inputs="I_CBF",
            output="BOLD",
            name=name, description=description
        )

#################################
### Two input channel model
#################################
class balloon_two_inputs(BoldModel):
    """
    BOLD model with two input signals (CBF-driving and CMRO2-driving) for the ballon model and non-linear BOLD equation with revised coefficients based on Buxton et al. (2004), Friston et al. (2000) and Stephan et al. (2007).
    """
    def __init__(self):

        # damped harmonic oscillators, gamma->spring coefficient, kappa->damping coefficient
        # CBF --> gamma from Friston
        # CMRO2 --> faster --> gamma=gamma_CBF*10 (therefore scaling of I_CMRO2 by (gamma_CMRO2 / gamma_CBF) --> if same input (I_CBF==I_CMRO2) CMRO2 and CBF same steady-state)
        # critical kappa --> kappa**2-4*gamma = 0 --> kappa=sqrt(4*gamma)
        # CBF underdamped for undershoot --> kappa = 0.6*sqrt(4*gamma)
        # CMRO2 critical --> kappa = sqrt(4*gamma)
        # after CBF and CMRO2 standard balloon model with revised coefficients, parameter values = Friston et al. (2000)
        parameters = """
            kappa_CBF   = 0.7650920556760059
            gamma_CBF   = 1/2.46
            kappa_CMRO2 = 4.032389192727559
            gamma_CMRO2 = 10/2.46
            phi_CBF     = 1.0
            phi_CMRO2   = 1.0
            E_0         = 0.34
            tau         = 0.98
            alpha       = 0.33
            V_0         = 0.02
            v_0         = 40.3
            TE          = 40/1000.
            epsilon     = 1
            r_0         = 25
            tau_out1    = 0
            tau_out2    = 20
            second      = 1000
        """
        equations = """
            # CBF input
            I_CBF               = sum(I_CBF)                                                   : init=0
            second*ds_CBF/dt    = phi_CBF * I_CBF - kappa_CBF * s_CBF - gamma_CBF * (f_in - 1) : init=0
            second*df_in/dt     = s_CBF                                                        : init=1, min=0.01

            # CMRO2 input
            I_CMRO2             = sum(I_CMRO2)                                                 : init=0
            second*ds_CMRO2/dt  = phi_CMRO2 * I_CMRO2 * (gamma_CMRO2 / gamma_CBF) - kappa_CMRO2 * s_CMRO2 - gamma_CMRO2 * (r - 1) : init=0
            second*dr/dt        = s_CMRO2                                                      : init=1, min=0.01

            dv                  = f_in - v**(1 / alpha)
            tau_out             = if dv>0: tau_out1 else: tau_out2
            f_out               = v**(1/alpha) + tau_out * dv / (tau + tau_out)              : init=1, min=0.01
            
            dq/dt               = (r - (q / v) * f_out)  / (second*tau)                        : init=1, min=0.01
            dv/dt               = dv / (tau + tau_out)  / second                             : init=1, min=0.01

            # Revised coefficients
            k_1                 = 4.3 * v_0 * E_0 * TE
            k_2                 = epsilon * r_0 * E_0 * TE
            k_3                 = 1 - epsilon

            # Non-linear equation
            BOLD                = V_0 * (k_1 * (1 - q) + k_2 * (1 - (q / v)) + k_3 * (1 - v))  : init=0
        """
        name = "BOLD model with two inputs"
        description = "BOLD model with two inputs (CBF-driving and CMRO2-driving). Combination of neurovascular coupling of Friston et al. (2000) and non-linear Balloon model with revised coefficients (Buxton et al, 1998, Stephan et al, 2007)"

        BoldModel.__init__(self, 
            parameters=parameters, 
            equations=equations, 
            inputs=['I_CBF', 'I_CMRO2'],
            output="BOLD",
            name=name, 
            description=description
        )


#################################
### Our current BOLD computation
#################################
class balloon_maith2021(BoldModel):
    """
    The balloon model as used in Maith et al. (2021).
    """
    def __init__(self):

        parameters = """
            second    = 1000.0
            phi   = 1.0
            kappa = 0.665
            gamma = 0.412
            E_0       = 0.3424
            tau     = 1.0368
            alpha     = 0.3215
            V_0       = 0.02
        """
        equations = """
            I_CBF          = sum(I_CBF)                                                 : init=0
            ds/dt          = (phi * I_CBF - kappa * s - gamma * (f_in - 1))/second      : init=0
            df_in/dt       = s / second                                                 : init=1, min=0.01

            E              = 1 - (1 - E_0)**(1 / f_in)                                  : init=0.3424
            dq/dt          = (f_in * E / E_0 - (q / v) * f_out)/(tau*second)            : init=1, min=0.01
            dv/dt          = (f_in - f_out)/(tau*second)                                : init=1, min=0.01
            f_out          = v**(1 / alpha)                                             : init=1, min=0.01

            k_1            = 7 * E_0
            k_2            = 2
            k_3            = 2 * E_0 - 0.2

            BOLD           = V_0 * (k_1 * (1 - q) + k_2 * (1 - (q / v)) + k_3 * (1 - v)) : init=0
        """
        name = "Maith2021 BOLD model"
        description = "BOLD computation from Maith et al. (2021)."

        BoldModel.__init__(self, 
            parameters=parameters, 
            equations=equations,  
            inputs="I_CBF",
            output="BOLD",
            name=name, description=description
        )


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

# Code which should be added prior to kernels
# (directly imported by CodeGenerator if needed)
additional_global_functions = ""

launch_config = {
    'init': """
        _threads_per_block = 192;
        unsigned int tmp_blocks = static_cast<unsigned int>(ceil(static_cast<double>(nb_synapses())/static_cast<double>(_threads_per_block)));
        _nb_blocks = std::min<unsigned int>(65535, tmp_blocks);

    #ifdef _DEBUG
        std::cout << "Initial kernel configuration: " << _nb_blocks << ", " << _threads_per_block << std::endl;
    #endif
""",
    'update': """
        if (nb_blocks != -1) {
            _nb_blocks = static_cast<unsigned int>(nb_blocks);
            _threads_per_block = threads_per_block;
        }else{
            _threads_per_block = threads_per_block;
            unsigned int tmp_blocks = static_cast<unsigned int>(ceil(static_cast<double>(nb_synapses())/static_cast<double>(_threads_per_block)));
            _nb_blocks = std::min<unsigned int>(65535, tmp_blocks);
        }

    #ifdef _DEBUG
        std::cout << "Updated kernel configuration: " << _nb_blocks << ", " << _threads_per_block << std::endl;
    #endif
"""
}

attribute_decl = {
    'local': """
    // Local %(attr_type)s %(name)s
    std::vector< %(type)s > %(name)s;
    %(type)s* gpu_%(name)s;
    long int %(name)s_device_to_host;
    bool %(name)s_host_to_device;
""",
    'semiglobal': """
    // Semiglobal %(attr_type)s %(name)s
    std::vector< %(type)s >  %(name)s ;
    %(type)s* gpu_%(name)s;
    long int %(name)s_device_to_host;
    bool %(name)s_host_to_device;
""",
    'global': """
    // Global %(attr_type)s %(name)s
    %(type)s %(name)s;
    %(type)s* gpu_%(name)s;
    long int %(name)s_device_to_host;
    bool %(name)s_host_to_device;
"""
}

attribute_cpp_init = {
    'local': """
        // Local %(attr_type)s %(name)s
        %(name)s = init_matrix_variable<%(type)s>(%(init)s);
        gpu_%(name)s = init_matrix_variable_gpu<%(type)s>(%(name)s);
        %(name)s_host_to_device = true;
        %(name)s_device_to_host = t;
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
        %(name)s = init_vector_variable<%(type)s>(%(init)s);
        gpu_%(name)s = init_vector_variable_gpu<%(type)s>(%(name)s);
        %(name)s_dirty = true;
""",
    'global': """
        // Global %(attr_type)s %(name)s
        %(name)s = static_cast<%(type)s>(%(init)s);
        cudaMalloc((void**)&gpu_%(name)s, sizeof(%(type)s));
        %(name)s_dirty = true;
"""
}

attribute_cpp_size = {
    'local': """
        // Local %(attr_type)s %(name)s
        size_in_bytes += sizeof(bool);
        size_in_bytes += sizeof(%(ctype)s*);
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
        size_in_bytes += sizeof(bool);
        size_in_bytes += sizeof(%(ctype)s*);
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'global': """
        // Global
        size_in_bytes += sizeof(bool);
        size_in_bytes += sizeof(%(ctype)s*);
        size_in_bytes += sizeof(%(ctype)s);
"""
}

attribute_cpp_delete = {
    'local': """
        // %(name)s
        cudaFree(gpu_%(name)s);
""",
    'semiglobal': """
        // %(name)s
        cudaFree(gpu_%(name)s);
""",
    'global': ""
}

attribute_host_to_device = {
    'local': """
        // %(name)s: local
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( gpu_%(name)s, %(name)s.data(), this->nb_synapses() * sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
""",
    'semiglobal': """
        // %(name)s: semiglobal
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( gpu_%(name)s, %(name)s.data(), this->nb_dendrites() * sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
""",
    'global': """
        // %(name)s: global
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( gpu_%(name)s, &%(name)s, sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
"""
}

attribute_device_to_host = {
    'local': """
        // %(name)s: local
        if ( %(name)s_device_to_host < t ) {
        #ifdef _DEBUG
            std::cout << "DtoH: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( %(name)s.data(), gpu_%(name)s, this->nb_synapses() * sizeof( %(type)s ), cudaMemcpyDeviceToHost);
        #ifdef _DEBUG
            cudaError_t err_%(name)s = cudaGetLastError();
            if ( err_%(name)s != cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err_%(name)s) << std::endl;
        #endif
            %(name)s_device_to_host = t;
        }
""",
    'semiglobal': """
        // %(name)s: semiglobal
        if ( %(name)s_device_to_host < t ) {
        #ifdef _DEBUG
            std::cout << "DtoH: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( %(name)s.data(), gpu_%(name)s, this->nb_dendrites() * sizeof(%(type)s), cudaMemcpyDeviceToHost);
        #ifdef _DEBUG
            cudaError_t err_%(name)s = cudaGetLastError();
            if ( err_%(name)s != cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err_%(name)s) << std::endl;
        #endif
            %(name)s_device_to_host = t;
        }
""",
    'global': """
        // %(name)s: global
        if ( %(name)s_device_to_host < t ) {
        #ifdef _DEBUG
            std::cout << "DtoH: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( &%(name)s, gpu_%(name)s, sizeof(%(type)s), cudaMemcpyDeviceToHost);
        #ifdef _DEBUG
            cudaError_t err_%(name)s = cudaGetLastError();
            if ( err_%(name)s != cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err_%(name)s) << std::endl;
        #endif
            %(name)s_device_to_host = t;
        }
"""
}

rate_psp_kernel = {
    'device_kernel': {
        'sum': """
__global__ void cu_proj%(id_proj)s_psp_coo(%(conn_args)s%(add_args)s, %(float_prec)s* %(target_arg)s ) {
    %(size_type)s j = segments[blockIdx.x] + threadIdx.x;
    %(size_type)s C = segments[blockIdx.x+1];
    %(size_type)s block_off = segment_size * blockIdx.x;

    extern %(float_prec)s __shared__ sdata[];

    if (threadIdx.x < segment_size)
        sdata[threadIdx.x] = 0.0;
    __syncthreads();

    for( ; j < C; j += blockDim.x ) {

        %(float_prec)s sum = %(psp)s
        %(idx_type)s loc_idx = row_indices[j]-block_off;
        atomicAdd(&(sdata[loc_idx]), sum);
    }

    __syncthreads();
    if (threadIdx.x < segment_size)
        %(target_arg)s[block_off + threadIdx.x] += sdata[threadIdx.x];
}
"""
    },
    'kernel_decl': """__global__ void cu_proj%(id)s_psp_coo(%(conn_args)s%(add_args)s, %(float_prec)s* %(target_arg)s );
""",
    'host_call': """
    // proj%(id_proj)s: pop%(id_pre)s -> pop%(id_post)s
    if ( pop%(id_post)s->_active && proj%(id_proj)s->_transmission ) {
        int sharedMemSize = proj%(id_proj)s.segment_size() * sizeof(%(float_prec)s);
        cu_proj%(id_proj)s_psp_coo<<< proj%(id_proj)s.number_of_segments(), proj%(id_proj)s->_threads_per_block, sharedMemSize >>>(
            %(conn_args)s
            /* other variables */
            %(add_args)s
            /* result */
            %(target_arg)s
        );

    #ifdef _DEBUG
        auto err = cudaGetLastError();
        if ( err != cudaSuccess ) {
            std::cout << "cu_proj%(id_proj)s_psp: " << cudaGetErrorString(err) << std::endl;
        }
    #endif
    }
""",
    'thread_init': {
        'float': {
            'sum': "0.0f",
            'min': "FLT_MAX",
            'max': "FLT_MIN",
            'mean': "0.0f"
        },
        'double': {
            'sum': "0.0",
            'min': "DBL_MAX",
            'max': "DBL_MIN",
            'mean': "0.0"
        }
    }
}

conn_templates = {
    # connectivity representation
    'conn_header': "const %(idx_type)s segment_size, const %(size_type)s *segments, const %(idx_type)s* __restrict__ row_indices, const %(idx_type)s* __restrict__ column_indices",
    'conn_call': "proj%(id_proj)s->segment_size(), proj%(id_proj)s->gpu_segments(), proj%(id_proj)s->gpu_row_indices(), proj%(id_proj)s->gpu_column_indices()",
    'conn_kernel': "",

    # launch config
    'launch_config': launch_config,

    # accessors
    'attribute_decl': attribute_decl,
    'attribute_cpp_init': attribute_cpp_init,
    'attribute_cpp_size': attribute_cpp_size,
    'attribute_cpp_delete': attribute_cpp_delete,
    'host_to_device': attribute_host_to_device,
    'device_to_host': attribute_device_to_host,

    # operations
    'rate_psp': rate_psp_kernel,
}

conn_ids = {
    'local_index': "[j]",
    'semiglobal_index': '[i]',
    'global_index': '[0]',
    'pre_index': '[column_indices[j]]',
    'post_index': '[row_indices[j]]',
    'pre_prefix': 'pre_',
    'post_prefix': 'post_',
}


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.core import Global
from ANNarchy.core.Network import Network
from ANNarchy.core.Population import Population
from ANNarchy.core.Projection import Projection
from ANNarchy.core.Monitor import Monitor
from ANNarchy.core.Random import Uniform
from ANNarchy.extensions.convolution import Convolution, Pooling
from ANNarchy.intern import Messages

import numpy as np

from .InputEncoding import CPN, IB, PSO
from .ReadOut import available_read_outs, IaF, IaF_ReadOut, IaF_TTKS, IaF_Acc

class ANNtoSNNConverter :
    r"""
    Converts a pre-trained Keras model `.keras` into an ANNarchy spiking neural network. 

    The implementation of the present module is inspired by the SNNToolbox (Rueckauer et al. 2017), and is largely based on the work of Diehl et al. (2015). We provide several input encodings, as suggested in the work of Park et al. (2019) and Auge et al. (2021).

    **Constraints on the ANN**

    It is not possible to convert any keras model into an ANNarchy SNN: some constraints ahve to be respected.

    * The only allowed layers in the ANN are:
        * `Dense`
        * `Conv2D`
        * `MaxPooling2D`
        * `AveragePooling2D`
        * as well as non-neural layers such as Dropout, Activation, BatchNorm, etc.

    * The layers must **not** contain any bias, even the conv layers:

    ```python
    x = tf.keras.layers.Dense(128, use_bias=False, activation='relu')(x)
    x = tf.keras.layers.Conv2D(64, kernel_size=(5,5), activation='relu', padding='same', use_bias=False)(x)
    ```

    * The first layer of the network must be an `Input`:

    ```python
    inputs = tf.keras.Input(shape = (28, 28, 1))
    ```

    * Pooling must explicitly be done by `MaxPooling2D`/`AveragePooling2D`, strides are ignored.
    
    Please be aware that the module is very experimental and the conversion may not work for many different reasons. Feel free to submit issues.

    **Processing Queue**

    The pre-trained ANN model to be converted should be saved in keras (extension `.keras`). The saved model is transformed layer by layer into a feed-forward ANNarchy spiking network. The structure of the network remains the same as in the original ANN, while the weights are normalised. Please note that the current implementation focuses primarily on the correctness of the conversion. Computational performance, especially of the converted CNNs, will be improved in future releases.

    :::callout-note
    
    While the neurons are conceptually spiking neurons, there is one specialty: next to the spike event (stored automatically in ANNarchy), each event will be stored in an additional *mask* array. This *mask* value decays in absence of further spike events exponentially. The decay can be controlled by the *mask_tau* parameter of the population. The projections (either dense or convolution) will use this mask as pre-synaptic input, not the generated list of spike events.
    :::

    **Input Encoding**

    * Poisson ("CPN")

    This encoding uses a Poisson distribution where the pixel values of the image will be used as probability for each individual neuron.
    
    * Intrinsically Bursting ("IB")

    This encoding is based on the Izhikevich (2003) model that comprises two ODEs:

    $$
    \begin{cases}
    \frac{dv}{dt} = 0.04 \cdot v^2 + 5.0 \cdot v + 140.0 - u + I \\
    \\
    \frac{du}{dt} = a \cdot (b \cdot v - u) \\
    \end{cases}
    $$

    The parameters for $a$ - $d$ are selected accordingly to Izhikevich (2003). The provided input images will be set as $I$.


    * Phase Shift Oscillation ("PSO")

    Based on the description by Park et al. (2019), the spiking threshold $v_\text{th}$ is modulated by a oscillation function $\Pi$, whereas the membrane potential follows simply the input current. 

    $$
    \begin{cases}
    \Pi(t) = 2^{-(1+ \text{mod}(t,k))}\\
    \\
    v_\text{th}(t) = \Pi(t) \, v_\text{th}(t)\\
    \end{cases}
    $$

    * User-defined input encodings

    In addition to the pre-defined models, one can opt for individual models using the `Neuron` class of ANNarchy. Please note that a `mask` variable need to be defined, which is fed into the subsequent projections.

    **Read-out Methods**

    In a classification task, the neuron with the highest activity corresponds corresponds to the decision to which class the presented input belongs. However, the highest activity can be determined in different ways. We support currently three methods, defined by the `read_out` parameter of the constructor:

    * Maximum Spike Count

    `read_out = 'spike_count'` : the number of spikes emitted by each neuron is recorded and the index of the neuron(s) with the maximum number is returned.

    * Time to Number of Spikes

    `read_out = 'time_to_first_spike'` or `read_out = 'time_to_k_spikes'`: when the first or first $k$ spikes are emitted by a single neuron, the simulation is stopped and the neuron rank(s) is returned. For the second mode, an additional $k$ argument need to be also provided.

    * Membrane potential

    `read_out = 'membrane_potential'`:  pre-synaptic events are accumulated in the membrane potential of each output neuron. The index of the neuron(s) with the highest membrane potential is returned.

    > Izhikevich (2003) Simple Model of Spiking Neurons. IEEE transactions on neural networks 14(6). doi: 10.1109/TNN.2003.820440

    > Diehl PU, Neil D, Binas J,et al. (2015) Fast-classifying, high-accuracy spiking deep networks through weight and threshold balancing, 2015 International Joint Conference on Neural Networks (IJCNN), 1-8, doi: 10.1109/IJCNN.2015.7280696.

    > Rueckauer B, Lungu I, Hu Y, et al. (2017) Conversion of Continuous-Valued Deep Networks to Efficient Event-Driven Networks for Image Classification., Front. Neurosci., 2017, 11. doi: 10.3389/fnins.2017.00682

    > Park S, Kim S, Choe H, et al. (2019) Fast and Efficient Information Transmission with Burst Spikes in Deep Spiking Neural Networks. 

    > Auge D, Hille J, Mueller E et al. (2021) A Survey of Encoding Techniques for Signal Processing in Spiking Neural Networks. Neural Processing Letters. 2021; 53:4963-4710. doi:10.1007/s11063-021-10562-2

    :param input_encoding: a string representing which input encoding should be used: 'CPN', 'IB' or 'PSO'.
    :param hidden_neuron:  neuron model used in the hidden layers. Either the default integrate-and-fire ('IaF') or an ANNarchy Neuron object.
    :param read_out: a string which of the following read-out method should be used: `spike_count`, `time_to_first_spike`, `membrane_potential`.
    """

    def __init__(self, 
                 input_encoding:str='CPN', 
                 hidden_neuron:str='IaF', 
                 read_out:str='spike_count', 
                 **kwargs):

        # Neuron model
        if isinstance(hidden_neuron, str):
            if hidden_neuron == "IaF":
                self._hidden_neuron_model = IaF
            else:
                raise ValueError("Invalid model name for hidden neurons.")
        else:
            self._hidden_neuron_model = hidden_neuron

        # Input encoding
        self._input_encoding = input_encoding
        if input_encoding == "poisson" or input_encoding == "CPN":
            self._input_model = CPN
        elif input_encoding == 'PSO':
            self._input_model = PSO
        elif input_encoding=='IB':
            self._input_model = IB
        else:
            raise ValueError("Unknown input encoding:", input_encoding)

        # Readout
        if read_out in available_read_outs:
            self._read_out = read_out
        else:
            raise ValueError("Unknown value for read-out:", read_out)

        # Maximum frequency
        self._max_f = 100      # scale factor used for poisson encoding

        if self._read_out == "time_to_k_spikes":
            if 'k' not in kwargs.keys():
                Messages._error("When read_out is set to 'time_to_k_spikes', the k parameter need to be provided.")
            self._k_param = kwargs['k']

        # TODO: sanity check on key-value args
        for key, value in kwargs.items():
            if key == "max_f":
                self._max_f = value

        self._snn_network = None

    def load_keras_model(self, 
            filename:str, 
            directory:str="annarchy", 
            scale_factor:float=None, 
            show_info:bool=True, 
        ) -> "Network":
        """
        Loads the pre-trained model provided as a .keras file.

        In tf.keras, the weights can be saved using:

        ```python
        model.save("model.keras")
        ```

        :param filename: path to the `.keras` file.
        :param directory: sub-directory where the generated code should be stored (default: "annarchy")
        :param scale_factor: allows a fine-grained control of the weight scale factor. By default (None), with each layer-depth the factor increases by one. If a scalar value is provided the same value is used for each layer. Otherwise a list can be provided to assign the scale factors individually.
        :param show_info: whether the network structure should be printed on console (default: True)
        """
        description = ""

        # Load keras model
        import tensorflow as tf
        model = tf.keras.models.load_model(filename)

        # Create spiking network
        self._snn_network = Network()

        # Input Population
        if not isinstance(model.layers[0], tf.keras.layers.InputLayer):
            Messages._error("The first layer of the network must be an Input layer.")
        
        input_name = model.layers[0].name
        input_shape = get_shape(model.layers[0].output)[1:]

        input_pop = self._snn_network.create(
            geometry=input_shape, 
            neuron=self._input_model,
            name = input_name, 
        )

        description += f"* Input layer: {input_name}, {input_shape}\n"

        # Iterate over layers
        weights = []
        
        for idx, layer in enumerate(model.layers):

            # Get the name
            name = layer.name

            # Is the layer legit in ANNarchy?
            pop = None

            # The only supported layers are dense, conv* and pool*
            if isinstance(layer, tf.keras.layers.Dense):
                
                # Get incoming weights
                W = layer.get_weights()[0].T
                size = W.shape[0]

                # Neuron type
                neuron_type = self._hidden_neuron_model
                stop_condition = None

                # Readout layer can have different neurons
                readout = False
                if idx == len(model.layers) - 1:
                    readout = True
                    if self._read_out == "membrane_potential":
                        neuron_type = IaF_Acc 
                        stop_condition = None
                    elif self._read_out == "time_to_first_spike":
                        neuron_type = IaF
                        stop_condition="spiked: any"
                    elif self._read_out == "time_to_k_spikes":
                        neuron_type = IaF_TTKS 
                        stop_condition="sc >= k: any"
                
                # Create the population
                pop = self._snn_network.create(
                    geometry = size, 
                    neuron=neuron_type, 
                    name=name,
                    stop_condition=stop_condition
                )

                if not readout:
                    #pop.vt = pop.vt - (0.05 * len(pops))
                    pop.vt = pop.vt - (0.05 * idx)

                description += f"* Dense layer: {name}, {size} \n"

                # Create projection
                proj = self._snn_network.connect(
                    pre = self._snn_network.get_populations()[-2], post = pop, 
                    target = "exc", 
                    name=f"dense_proj_{len(self._snn_network.get_populations())}"
                )

                proj.from_matrix(W, storage_format="dense")

                # Add to the network
                weights.append(W)

                description += f"    weights: {W.shape}\n"
                description += f"    mean {np.mean(W)}, std {np.std(W)}\n"
                description += f"    min {np.min(W)}, max {np.max(W)}\n"
                

            elif isinstance(layer, tf.keras.layers.Conv2D):
                
                geometry = get_shape(layer.output)[1:]

                pop = self._snn_network.create(
                    geometry = geometry, 
                    neuron=self._hidden_neuron_model, 
                    name=name
                ) 
                
                #pop.vt = pop.vt - (0.05 * len(pops))
                pop.vt = pop.vt - (0.05 * idx)

                description += f"* Conv2D layer: {name}, {geometry} \n"

                W = layer.get_weights()[0]
                W = np.moveaxis(W, -1, 0)

                proj = self._snn_network.connect(
                    Convolution(
                        pre = self._snn_network.get_populations()[-2], post = pop, target='exc', 
                        psp="pre.mask * w", 
                        name=f'conv_proj_{len(self._snn_network.get_populations())}'
                    )
                )
                proj.connect_filters(weights=W)

                weights.append(W)

            elif isinstance(
                layer, 
                (tf.keras.layers.MaxPooling2D,
                 tf.keras.layers.AveragePooling2D)
                ):
                
                geometry = get_shape(layer.output)[1:]
                pool_size = layer.get_config()['pool_size'] + (1,)
                operation = 'max' if isinstance(layer, tf.keras.layers.MaxPooling2D) else 'mean'

                pop = self._snn_network.create(
                    geometry = geometry, 
                    neuron=self._hidden_neuron_model, 
                    name=name
                ) 
                
                #pop.vt = pop.vt - (0.05 * len(pops))
                pop.vt = pop.vt - (0.05 * idx)

                description += f"* MaxPooling2D layer: {name}, {geometry} \n"
                
                proj = self._snn_network.connect(
                    Pooling(
                        pre = self._snn_network.get_populations()[-2], post = pop, target='exc', 
                        operation=operation, psp="pre.mask", 
                        name=f'pool_proj_{len(self._snn_network.get_populations())}'
                    )
                )
                
                proj.pooling(extent=pool_size)
                weights.append([])

            # Compatible layer has not been found
            if pop is None:
                description += f"* {type(layer).__name__} skipped.\n"

        # Record the last layer to determine prediction
        self._monitor = self._snn_network.monitor(pop, ['v', 'spike'])

        # Compile the configured network
        self._snn_network.compile(directory=directory, silent=True)

        if show_info:
            print(description)

        # Normalize the weights
        factors = self._normalize_weights(weights, scale_factor)
        for i in range(len(weights)):
            if len(weights[i]) > 0:
                self._snn_network.get_populations()[i].weights = weights[i] * float(factors[i])

        return self._snn_network


    def get_annarchy_network(self) -> "Network":
        """
        Returns the ANNarchy.Network instance.
        """
        return self._snn_network

    def predict(self, 
                samples:np.ndarray, 
                duration_per_sample:int=1000, 
                multiple:bool=False,
                ) -> list[int]:
        """
        Performs the prediction for a given input array.

        :param samples: set of inputs to present to the network. The function expects a 2-dimensional array (num_samples, input_size).
        :param duration_per_sample: the number of simulation steps for one input sample (default: 1000, 1 second biological time)
        :param multiple: if several output neurons reach the criteria, return the full list instead of randomly chosing one.
        """

        predictions = []
        nb_classes = self._snn_network.get_populations()[-1].size

        # Use the progress bar
        try:
            import tqdm
        except Exception as e:
            iterator = range(samples.shape[0])
        else:
            iterator = tqdm.tqdm(range(samples.shape[0]))

        # Iterate over all samples
        for i in iterator:

            # Reset state variables
            self._snn_network.reset(populations=True, monitors=True, projections=False)

            # Set input
            self._snn_network.get_populations()[0].rates =  samples[i,:] * self._max_f

            # The read-out is performed differently based on the mode selected by the user
            if self._read_out in ["time_to_first_spike", "time_to_k_spikes"]:
                # Simulate until the condition is met
                self._snn_network.simulate_until(duration_per_sample, population=self._snn_network.get_populations()[-1])

                # Read-out accumulated inputs
                data = self._monitor.get('spike')
                act_pred = np.zeros(nb_classes)
                for neur_rank, spike_times in data.items():
                    act_pred[neur_rank] = len(spike_times)

                # Gather all neurons which fulfilled the condition
                prediction = np.argwhere(act_pred == np.amax(act_pred)).flatten()

            elif self._read_out == "membrane_potential":
                # Simulate 1s and record spikes in output layer
                self._snn_network.simulate(duration_per_sample)

                # Read-out accumulated inputs
                data = self._monitor.get('v')

                # The neuron with the highest accumulated membrane potential is the selected candidate
                prediction = np.argwhere(data[-1,:] == np.amax(data[-1,:])).flatten()

            elif self._read_out == "spike_count":
                # Simulate 1s and record spikes in output layer
                self._snn_network.simulate(duration_per_sample)

                # Retrieve the recorded spike events
                data = self._monitor.get('spike')

                # The predicted label is the neuron index with the highest number of spikes.
                # Therefore, we count the number of spikes each output neuron emitted.
                act_pred = np.zeros(nb_classes)
                for neur_rank, spike_times in data.items():
                    act_pred[neur_rank] = len(spike_times)

                # Gather all neurons which achieved highest number of spikes
                prediction = np.argwhere(act_pred == np.amax(act_pred)).flatten()

            else:
                raise NotImplementedError
            
            # Treat ex-aequo
            if multiple:
                predictions.append(list(prediction))
            else:
                predictions.append(int(np.random.choice(prediction, 1)))

        return predictions


    def _normalize_weights(self, weight_matrices, scale_factor=None):
        """
        Weight normalization based on the "model based normalization" from Diehl et al. (2015)
        """

        # Scale factor increases for each layer
        if scale_factor is None:
            scale_factor = np.arange(2, len(weight_matrices)+2)
        elif isinstance(scale_factor, (float, int)):
            scale_factor = [scale_factor] * len(weight_matrices)
        elif isinstance(scale_factor, (list, np.ndarray)):
            if len(scale_factor) != len(weight_matrices):
                Messages._error("The length of the scale_factor list must be equal to the number of projections.")
            else:
                pass # nothing to do
        else:
            raise ValueError("Invalid argument for scale_factor", type(scale_factor))

        # Iterate over all weight matrices
        factors = []
        for level, w_matrix in enumerate(weight_matrices):

            factor = 1.0

            if len(w_matrix)> 0: # Empty weight matrix is for max-pooling

                # Reshape the matrix with post-neurons first
                w = w_matrix.reshape((w_matrix.shape[0], -1))

                # Maximum input current over all post neurons
                max_val = np.sum(w * (w > 0), axis=1).max()

                # Normalize the incoming weights for each neuron, based on the maximum input for the complete connection 
                # and multiply it with the depth of the connection to boost the input current
                factor = scale_factor[level]  / max_val

            factors.append(factor)

        return factors

def get_shape(tensor) -> tuple:
    """
    Returns the shape of the tensorflow tensor as a tuple.

    tf < 2.14 returned a TensorShape object, but now it is a tuple.
    
    """
    if isinstance(tensor.shape, (tuple,)):
        return tensor.shape
    else:
        try:
            return tuple(tensor.shape.as_list())
        except:
            Messages._error("ANN_to_SNN: unable to estimate the layer's size.")




"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.intern.ConfigManagement import ConfigManager
from ANNarchy.intern.GlobalObjects import GlobalObjectManager
from ANNarchy.intern import Messages
from ANNarchy.parser.AnalyseSynapse import analyse_synapse

class Synapse :
    """
    Base class to define a synapse model.

    Synapses expect `parameters` as a dictionary and `equations` as a list of variable updates (including `w` if there is synaptic plasticity). 

    Rate-coded synapses can define `psp` and `operation` to modify synaptic transmission:

    ```python
    nonlinear_synapse = ann.Synapse( 
        psp = "log( (pre.r * w + 1 ) / (pre.r * w - 1) )",
        operation = 'max',
    )
    ```

    Spiking synapses can define event-based rules, such as `pre_spike` (a pre-synaptic spike arrives at the synapse) and `post_spike` (the post-synaptic neuron emits a spike):

    ```python
    STDP = ann.Synapse(
        parameters = dict(
            tau_pre = 10.0,
            tau_post = 10.0,
            cApre = 0.01,
            cApost = 0.0105,
            wmax = 0.01,
        ),
        equations = [
            ann.Variable('tau_pre * dApre/dt = - Apre', method='event-driven'),
            ann.Variable('tau_post * dApost/dt = - Apost', method='event-driven'),
        ],
        pre_spike = '''
            g_target += w
            Apre += cApre * wmax
            w = clip(w - Apost, 0.0 , wmax)
        ''',                  
        post_spike = '''
            Apost += cApost * wmax
            w = clip(w + Apre, 0.0 , wmax)
        ''' 
    )
    ```

    :param parameters: dictionary of parameters and their initial value.
    :param equations: list of equations defining the temporal evolution of variables.
    :param psp: continuous influence of a single synapse on the post-synaptic neuron (default for rate-coded: ``w*pre.r``). Synaptic transmission in spiking synapses occurs in ``pre_spike``.
    :param operation: operation (sum, max, min, mean) performed by the post-synaptic neuron on the individual psp (rate-coded only).
    :param pre_spike: updating of variables when a pre-synaptic spike is received (spiking only).
    :param post_spike: updating of variables when a post-synaptic spike is emitted (spiking only).
    :param pre_axon_spike: updating of variables when an axonal spike was emitted (spiking only, default None). The usage of this arguments prevents the application of learning rules.
    :param functions: additional functions used in the equations.
    :param pruning: Condition for pruning the synapse.
    :param creating: Condition for creating the synapse.
    :param name: name of the synapse type (used for reporting only).
    :param description: short description of the synapse type (used for reporting).
    """

    # Default name and description for reporting
    _default_names = {'rate': "Rate-coded synapse", 'spike': "Spiking synapse"}

    def __init__(self, 
                 parameters:str|dict="", 
                 equations:str|list="", 
                 psp:str=None, 
                 operation:str='sum', 
                 pre_spike:str|list=None, 
                 post_spike:str|list=None, 
                 pre_axon_spike:str=None, 
                 functions:str=None, 
                 pruning:str=None, 
                 creating:str=None, 
                 name:str=None, 
                 description:str=None, 
                 extra_values:dict={} ):


        # Store the parameters and equations
        self.parameters = parameters
        self.equations = equations
        self.functions = functions
        self.pre_spike = pre_spike
        self.post_spike = post_spike
        self.psp = psp
        self.pre_axon_spike = pre_axon_spike
        self.operation = operation
        self.extra_values = extra_values
        self.pruning = pruning
        self.creating = creating

        # Type of the synapse TODO: smarter
        self.type = 'spike' if pre_spike else 'rate'

        # Check the operation
        if self.type == 'spike' and self.operation != 'sum':
            Messages._error('Spiking synapses can only perform a sum of presynaptic potentials.')

        if not self.operation in ['sum', 'min', 'max', 'mean']:
            Messages._error('The only operations permitted are: sum (default), min, max, mean.')

        # Sanity check
        if self.pre_axon_spike and self.post_spike:
            Messages._error("The usage of axonal spike events is currently not allowed for plastic connections.")


        # Description
        self.description = None

        # Reporting
        if not hasattr(self, '_instantiated') : # User-defined
            GlobalObjectManager().add_synapse_type(synapse=self)
        elif len(self._instantiated) == 0: # First instantiation of the class
            GlobalObjectManager().add_synapse_type(synapse=self)
        self._rk_synapses_type = GlobalObjectManager().num_synapse_types()

        if name:
            self.name = name
        else:
            self.name = self._default_names[self.type]

        if description:
            self.short_description = description
        else:
            if self.type == 'spike':
                self.short_description = "User-defined spiking synapse."
            else:
                self.short_description = "User-defined rate-coded synapse."

    def _analyse(self, net_id):
        # Analyse the synapse type
        if not self.description:
            self.description = analyse_synapse(self, net_id)

    def __add__(self, synapse):
        Messages._error('adding synapse models is not implemented yet.')

        #self._variables.update(synapse.variables)

    def __repr__(self):
        if self.type == 'rate':
            text= """Rate-coded synapse.

Parameters:
""" + str(self.parameters) + """
Equations of the variables:
""" + str(self.equations) + """

""" + """
Synaptic transmission (psp):

""" + "\tw*pre.r" if self.psp == None else str(self.psp)

        else:
            text= """Spiking synapse.

Parameters:
""" + str(self.parameters) + """
Equations of the variables:
""" + str(self.equations) + """
pre-synaptic spike:
""" + str(self.pre_spike) + """
post-synaptic spike:
""" + str(self.post_spike)

        return text

    def __str__(self):
        import pprint
        return pprint.pformat( self, depth=4 ) #TODO
